
import datetime as dt

import numpy as np
import pandas as pd
from bokeh.layouts import column, row
from bokeh.models import Div

from .data_frame import interpolate_to
from .plot import dot_map, line_diff, cumulative, heart_rate_zones, health, line_diff_elevation_climbs, \
    max_all, min_all, activities
from .server import Page
from ..data import statistics
from ..data.data_frame import set_log, activity_statistics
from ..lib.date import format_time, format_seconds
from ..squeal import ActivityGroup
from ..squeal import ActivityJournal
from ..stoats.calculate.monitor import MonitorStatistics
from ..stoats.names import SPEED, DISTANCE, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, TIME, FATIGUE, FITNESS, \
    ACTIVE_DISTANCE, ACTIVE_TIME, HR_ZONE, ELEVATION, REST_HR, DAILY_STEPS, CLIMB_ELEVATION, CLIMB_DISTANCE, ALTITUDE

WINDOW = '60s'
#WINDOW = 10
MIN_PERIODS = 1

DISTANCE_KM = '%s / km' % DISTANCE
SPEED_KPH = '%s / kph' % SPEED
MED_SPEED_KPH = 'M(%s) %s / kph' % (WINDOW, SPEED)
ELEVATION_M = '%s / m' % ELEVATION
CLIMB_MPS = 'Climb / mps'
HR_10 = 'HR Impulse / 10s'
MED_HR_10 = 'M(%s) HR Impulse / 10s' % WINDOW

LOG_FITNESS = 'Log %s' % FITNESS
LOG_FATIGUE = 'Log %s' % FATIGUE

RIDE_PLOT_LEN = 700
RIDE_PLOT_HGT = 200
HEALTH_PLT_LEN = 500
HEALTH_PLOT_HGT = 200
MAP_LEN = 400


def get(df, name):
    if name in df:
        return df[name]
    else:
        return None


def comparison(log, s, activity, compare=None):

    # ---- definitions

    set_log(log)
    names = [SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DISTANCE, ELEVATION, SPEED, HR_ZONE, HR_10, ALTITUDE]

    # ---- load data

    def get_stats(aj):
        st = [df for id, df in
              activity_statistics(s, *names, activity_journal_id=aj.id, with_timespan=True).groupby('timespan_id')]
        for df in st:
            df[DISTANCE_KM] = df[DISTANCE]/1000
            df[SPEED_KPH] = df[SPEED] * 3.6
            df[MED_SPEED_KPH] = df[SPEED].rolling(WINDOW, min_periods=MIN_PERIODS).median() * 3.6
            df[MED_HR_10] = df[HR_10].rolling(WINDOW, min_periods=1).median()
            df.rename(columns={ELEVATION: ELEVATION_M}, inplace=True)
        st_10 = [interpolate_to(df, HR_10) for df in st]
        for df in st_10:
            df[CLIMB_MPS] = df[ELEVATION_M].diff() * 0.1
        return st, st_10

    st1, st1_10 = get_stats(activity)
    st2, st2_10 = get_stats(compare) if compare else (None, None)
    climbs = activity_statistics(s, CLIMB_DISTANCE, CLIMB_ELEVATION, activity_journal_id=activity.id)
    line_x_range = None

    def all_frames(st, name):
        return [df[name].copy() for df in st]

    def set_axis(ys, st, name):
        if name != TIME:
            for y, df in zip(ys, st):
                y.index = df[name].copy()

    # ---- ride-specific plots

    def set_axes(y_axis, x_axis=TIME):
        y1 = all_frames(st1_10, y_axis)
        set_axis(y1, st1_10, x_axis)
        if compare:
            y2 = all_frames(st2_10, y_axis)
            set_axis(y2, st2_10, x_axis)
        else:
            y2 = None
        return y1, y2

    def ride_line(y_axis, x_axis=TIME):
        y1, y2 = set_axes(y_axis, x_axis=x_axis)
        return line_diff(RIDE_PLOT_LEN, RIDE_PLOT_HGT, x_axis, y1, y2, x_range=line_x_range)

    def ride_elevn(x_axis=TIME):
        y1, y2 = set_axes(ELEVATION_M, x_axis=x_axis)
        y3, _ = set_axes(ALTITUDE, x_axis=x_axis)
        return line_diff_elevation_climbs(RIDE_PLOT_LEN, RIDE_PLOT_HGT, y1, y2, climbs=climbs, st=st1, y3=y3,
                                          x_range=line_x_range)

    def ride_cum(y_axis):
        y1 = all_frames(st1_10, y_axis)
        if compare:
            y2 = all_frames(st2_10, y_axis)
        else:
            y2 = None
        return cumulative(RIDE_PLOT_HGT, RIDE_PLOT_HGT, y1, y2)

    hr10_line, hr10_cumulative = ride_line(MED_HR_10, x_axis=DISTANCE_KM), ride_cum(HR_10)
    line_x_range = hr10_line.x_range
    elvn_line, elvn_cumulative = ride_elevn(x_axis=DISTANCE_KM), ride_cum(CLIMB_MPS)
    speed_line, speed_cumulative = ride_line(MED_SPEED_KPH, x_axis=DISTANCE_KM), ride_cum(SPEED_KPH)

    mx, mn = max_all(df[HR_10] for df in st1_10), min_all(df[HR_10] for df in st1_10),
    for df in st1_10:
        df['size'] = MAP_LEN * ((df[HR_10] - mn) / (mx - mn)) ** 3 / 10
    x1, y1 = all_frames(st1_10, SPHERICAL_MERCATOR_X), all_frames(st1_10, SPHERICAL_MERCATOR_Y)
    if compare:
        x2, y2 = all_frames(st2_10, SPHERICAL_MERCATOR_X), all_frames(st2_10, SPHERICAL_MERCATOR_Y)
    else:
        x2, y2 = None, None
    map = dot_map(MAP_LEN, x1, y1, [df['size'] for df in st1_10], x2, y2)

    caption = '<table><tr><th>Name</th><th>Date</th><th>Duration</th><th>Distance</th></tr>'
    source_ids = [activity.id]
    if compare:
        source_ids.append(compare.id)
    st = statistics(s, ACTIVE_TIME, ACTIVE_DISTANCE, source_ids=source_ids)
    st_aj = st.loc[st.index == activity.start]
    caption += '<tr><td>%s</td><td>%s</td><td>%s</td><td>%.3f km</td></tr>' % \
               (activity.name, format_time(activity.start),
                format_seconds(st_aj[ACTIVE_TIME][0]), st_aj[ACTIVE_DISTANCE][0] / 1000)
    if compare:
        st_aj = st.loc[st.index == compare.start]
        caption += '<tr><td>%s</td><td>%s</td><td>%s</td><td>%.3f km</td></tr>' % \
                   (compare.name, format_time(compare.start),
                    format_seconds(st_aj[ACTIVE_TIME][0]), st_aj[ACTIVE_DISTANCE][0] / 1000)
    caption = Div(text=caption + '</table>', width=RIDE_PLOT_LEN, height=RIDE_PLOT_HGT)

    hrz_histogram = heart_rate_zones(RIDE_PLOT_HGT, RIDE_PLOT_HGT, pd.concat([df[HR_ZONE] for df in st1_10]))

    # ---- health-specific plots

    finish = activity.finish + dt.timedelta(days=1)  # to show new level
    start = finish - dt.timedelta(days=365)

    st_ff = statistics(s, FITNESS, FATIGUE, DAILY_STEPS, start=start, finish=finish)
    st_hr = statistics(s, REST_HR, owner=MonitorStatistics, start=start, finish=finish)
    st_ff[LOG_FITNESS] = np.log10(st_ff[FITNESS])
    st_ff[LOG_FATIGUE] = np.log10(st_ff[FATIGUE])

    health_line = health(HEALTH_PLT_LEN, HEALTH_PLOT_HGT, st_ff[LOG_FITNESS], st_ff[LOG_FATIGUE], st_hr[REST_HR])

    ajs = s.query(ActivityJournal). \
        join(ActivityGroup). \
        filter(ActivityJournal.start >= start,
               ActivityJournal.start < finish,
               ActivityGroup.id == activity.activity_group_id
               ).all()
    st_ac = statistics(s, ACTIVE_TIME, ACTIVE_DISTANCE, source_ids=[aj.id for aj in ajs])

    activity_line = activities(HEALTH_PLT_LEN, HEALTH_PLOT_HGT, get(st_ff, DAILY_STEPS), st_ac[ACTIVE_TIME])

    # ---- the final mosaic of plots

    return column(row(hr10_line, hr10_cumulative),
                  row(elvn_line, elvn_cumulative),
                  row(speed_line, speed_cumulative),
                  row(caption, hrz_histogram),
                  row(column(health_line, activity_line), map))


class ActivityJournalPage(Page):

    def create(self, s, id=None, compare=None, **kargs):
        aj1 = s.query(ActivityJournal).filter(ActivityJournal.id ==
                                              self.single_int_param('id', id)).one()
        title = aj1.name
        if compare:
            aj2 = s.query(ActivityJournal).filter(ActivityJournal.id ==
                                                  self.single_int_param('compare', compare)).one()
            title += ' v ' + aj2.name
        else:
            aj2 = None
        return {'header': title, 'title': title}, comparison(self._log, s, aj1, aj2)




from collections import defaultdict

import pandas as pd
from sqlalchemy import inspect, select, and_
from sqlalchemy.sql.functions import coalesce

from ch2.lib.data import MutableAttr, kargs_to_attr
from ch2.squeal import ActivityJournal
from ..squeal import StatisticName, StatisticJournal, StatisticJournalInteger, \
    StatisticJournalFloat, StatisticJournalText, Interval, StatisticMeasure, Source
from ..squeal.database import connect, ActivityTimespan

# because this is intended to be called from jupyter we hide the log here
# other callers can use these routines by calling set_log() first.
LOG = [None]


def df(query):
    # https://stackoverflow.com/questions/29525808/sqlalchemy-orm-conversion-to-pandas-dataframe
    return pd.read_sql(query.statement, query.session.bind)


def session(*args):
    ns, log, db = connect(args)
    set_log(log)
    return db.session()


def get_log():
    if not LOG[0]:
        raise Exception('Create session first (or call set_log from within python)')
    return LOG[0]


def set_log(log):
    LOG[0] = log


def _collect_statistics(s, names):
    if not names:
        names = ['%']
    statistic_ids, statistic_names = set(), set()
    for name in names:
        for statistic_name in s.query(StatisticName). \
                filter(StatisticName.name.like(name)).all():
            statistic_ids.add(statistic_name.id)
            statistic_names.add(statistic_name.name)
    return statistic_names, statistic_ids


def _tables():
    return kargs_to_attr(sj=inspect(StatisticJournal).local_table,
                         sn=inspect(StatisticName).local_table,
                         sji=inspect(StatisticJournalInteger).local_table,
                         sjf=inspect(StatisticJournalFloat).local_table,
                         sjt=inspect(StatisticJournalText).local_table,
                         inv=inspect(Interval).local_table,
                         at=inspect(ActivityTimespan).local_table)


def _build_statistic_journal_query(statistic_ids, start, finish, owner, constraint, source_ids, schedule):

    # use more efficient expression interface and exploit the fact that
    # alternative journal types will be null in an outer join.

    t = _tables()

    q = select([t.sn.c.name, t.sj.c.time, coalesce(t.sjf.c.value, t.sji.c.value, t.sjt.c.value)]). \
        select_from(t.sj.join(t.sn).outerjoin(t.sjf).outerjoin(t.sji).outerjoin(t.sjt)). \
        where(t.sn.c.id.in_(statistic_ids))
    if start:
        q = q.where(t.sj.c.time >= start)
    if finish:
        q = q.where(t.sj.c.time <= finish)
    if owner:
        q = q.where(t.sn.c.owner == owner)
    if constraint:
        q = q.where(t.sn.c.constraint == constraint)
    if source_ids is not None:  # avoid testing DataFrame sequences as bools
        # int() to convert numpy types
        q = q.where(t.sj.c.source_id.in_(int(id) for id in source_ids))
    if schedule:
        q = q.join(t.inv).where(t.inv.c.schedule == schedule)
    q = q.order_by(t.sj.c.time)
    return q


class MissingData(Exception): pass


def make_pad(data, times, statistic_names):
    err_cnt = defaultdict(lambda: 0)

    def pad():
        nonlocal data, times
        n = len(times)
        for name in statistic_names:
            if len(data[name]) != n:
                err_cnt[name] += 1
                if err_cnt[name] <= 1:
                    get_log().warning('Missing %s at %s (single warning)' % (name, times[-1]))
                data[name].append(None)

    return pad


def statistics(s, *statistics,
               start=None, finish=None, owner=None, constraint=None, source_ids=None, schedule=None):
    statistic_names, statistic_ids = _collect_statistics(s, statistics)
    q = _build_statistic_journal_query(statistic_ids, start, finish, owner, constraint, source_ids, schedule)
    data, times = defaultdict(list), []
    pad = make_pad(data, times, statistic_names)

    for name, time, value in s.connection().execute(q):
        if times and times[-1] != time:
            pad()
        if not times or times[-1] != time:
            times.append(time)
        if len(data[name]) >= len(times):
            raise Exception('Duplicate data for %s at %s ' % (name, time) +
                            '(you may need to specify more constraints to make the query unique)')
        data[name].append(value)
    pad()
    return pd.DataFrame(data, index=times)


def _resolve_activity(s, time, activity_journal_id):
    if activity_journal_id:
        if time:
            raise Exception('Specify activity_journal_id or time (not both)')
    else:
        if not time:
            raise Exception('Specify activity_journal_id or time')
        activity_journal_id = s.query(ActivityJournal.id). \
            filter(ActivityJournal.start <= time,
                   ActivityJournal.finish >= time).scalar()
        get_log().info('Using activity_journal_id=%d' % activity_journal_id)
    return activity_journal_id


def activity_statistics(s, *statistics,
                        time=None, activity_journal_id=None, with_timespan=False):

    statistic_names, statistic_ids = _collect_statistics(s, statistics)
    get_log().debug('Statistics IDs %s' % statistic_ids)
    activity_journal_id = _resolve_activity(s, time, activity_journal_id)

    t = _tables()
    q = select([t.sn.c.name, t.sj.c.time, coalesce(t.sjf.c.value, t.sji.c.value, t.sjt.c.value), t.at.c.id]). \
        select_from(t.sj.join(t.sn).outerjoin(t.sjf).outerjoin(t.sji).outerjoin(t.sjt)). \
        where(and_(t.sn.c.id.in_(statistic_ids), t.sj.c.source_id == activity_journal_id)). \
        select_from(t.at). \
        where(and_(t.at.c.start <= t.sj.c.time, t.at.c.finish >= t.sj.c.time,
                   t.at.c.activity_journal_id == activity_journal_id)). \
        order_by(t.sj.c.time)
    get_log().debug(q)

    data, times = defaultdict(list), []
    pad = make_pad(data, times, statistic_names)

    for name, time, value, timespan in s.connection().execute(q):
        if times and times[-1] != time:
            pad()
        if not times or times[-1] != time:
            times.append(time)
            if with_timespan:
                data['timespan_id'].append(timespan)
        if len(data[name]) >= len(times):
            raise Exception('Duplicate data for %s at %s ' % (name, time) +
                            '(you may need to specify more constraints to make the query unique)')
        data[name].append(value)
    pad()
    return pd.DataFrame(data, index=times)


def statistic_quartiles(s, *statistics,
                        start=None, finish=None, owner=None, constraint=None, source_ids=None, schedule=None):
    statistic_names, statistic_ids = _collect_statistics(s, statistics)
    q = s.query(StatisticMeasure). \
        join(StatisticJournal, StatisticMeasure.statistic_journal_id == StatisticJournal.id). \
        join(StatisticName, StatisticJournal.statistic_name_id == StatisticName.id). \
        join(Source, StatisticJournal.source_id == Source.id). \
        filter(StatisticName.id.in_(statistic_ids)). \
        filter(StatisticMeasure.quartile != None)
    if start:
        q = q.filter(StatisticJournal.time >= start)
    if finish:
        q = q.filter(StatisticJournal.time <= finish)
    if owner:
        q = q.filter(StatisticName.owner == owner)
    if constraint:
        q = q.filter(StatisticName.constraint == constraint)
    if source_ids is not None:
        q = q.filter(StatisticJournal.source_id.in_(source_ids))
    if schedule:
        q = q.join((Interval, StatisticMeasure.source_id == Interval.id)). \
            filter(Interval.schedule == schedule)
    get_log().debug(q)

    raw_data = defaultdict(lambda: defaultdict(lambda: [0] * 5))
    for measure in q.all():
        raw_data[measure.source.start][measure.statistic_journal.statistic_name.name][measure.quartile] = \
            measure.statistic_journal.value
    data, times = defaultdict(list), []
    for time in sorted(raw_data.keys()):
        times.append(time)
        sub_data = raw_data[time]
        for statistic in statistic_names:
            if statistic in sub_data:
                data[statistic].append(sub_data[statistic])
            else:
                data[statistic].append(None)

    return pd.DataFrame(data, index=times)


if __name__ == '__main__':
    s = session('-v 5')
    df = activity_statistics(s, 'Speed', time='2019-01-05 14:40:00', with_timespan=True)
    print(df.describe)




from abc import abstractmethod, ABC
from threading import Thread

from bokeh.server.server import Server
from tornado.ioloop import IOLoop

SINGLETON = None

TEMPLATE = '''
{% block css_resources %}
  {{ super() }}
  <style>
body {
  background-color: white;
}
.centre {
  text-align: center
}
.centre > div {
  display: inline-block;
}
table {
  margin: 20px;
  border-spacing: 10px;
}
  </style>
{% endblock %}
{% block inner_body %}
  <div class='centre'>
  <h1>{{ header }}</h1>
  {{ super() }}
  </div>
{% endblock %}
'''


class ServerThread(Server):

    def __init__(self, log, app_map):
        self._log = log
        super().__init__(app_map, io_loop=IOLoop.instance())
        Thread(target=self.io_loop.start).start()

    def stop(self):

        def callback():
            self._log.info('Shutting down server')
            self.io_loop.stop()
            super(ServerThread, self).stop(wait=True)
            self.unlisten()
            self._log.debug('Server shut down')

        self.io_loop.add_callback(callback)


def singleton_server(log, app_map):

    global SINGLETON

    if SINGLETON is None:
        log.info('Starting new server')
        # tornado is weird about ioloops and threads
        # https://github.com/universe-proton/universe-topology/issues/17
        SINGLETON = ServerThread(log, app_map)

    return SINGLETON


class Page(ABC):

    def __init__(self, log, db, template=TEMPLATE, **vars):
        self._log = log
        self._db = db
        self._template = template
        self._vars = vars

    def __call__(self, doc):
        self._log.debug('Page called with %s' % doc)
        with self._db.session_context() as s:
            params = doc.session_context.request.arguments
            self._log.debug('Request params: %s' % params)
            vars, root = self.create(s, **params)
            vars = vars if vars is not None else {}
            vars.update(self._vars)
            doc.add_root(root)
            doc.template = self._template
            doc.template_variables.update(vars)
            doc.title = vars['title'] if 'title' in vars else 'Choochoo'

    __ERROR = object()

    def single_int_param(self, name, values, deflt=__ERROR):
        try:
            return int(values[0])
        except Exception as e:
            msg = 'Could not parse "%s" for %s' % (values, name)
            self._log.warning(msg)
            if deflt != self.__ERROR:
                return deflt
            else:
                raise Exception(msg)

    @abstractmethod
    def create(self, s, **kargs):
        raise NotImplementedError()




from random import uniform
from unittest import TestCase

import pandas as pd
from bokeh.layouts import column, row
from bokeh.models import Label

from ch2.bucket.activity import DISTANCE_KM, ELEVATION_M, TEMPLATE
from ch2.bucket.plot import line_diff_elevation_climbs
from ch2.bucket.server import SingleShotServer
from ch2.config import config
from ch2.data.data_frame import _resolve_activity, set_log, activity_statistics
from ch2.lib.data import MutableAttr
from ch2.squeal import ActivityJournal
from ch2.stoats.calculate.climb import find_climbs, Climb
from ch2.stoats.names import DISTANCE, ELEVATION, CLIMB_ELEVATION, CLIMB_DISTANCE
from ch2.stoats.read.segment import SegmentImporter
from ch2.stoats.waypoint import WaypointReader


class TestClimb(TestCase):

    def build_climb(self, xys, speed=10, noise=0, dt=1):
        t0 = 0
        x0, y0 = xys[0]
        t, x, y = t0, x0, y0
        for x1, y1 in xys[1:]:
            t1 = x1 / speed
            while t <= t1:
                x = x0 + (x1 - x0) * (t - t0) / (t1 - t0)
                y = y0 + (y1 - y0) * (t - t0) / (t1 - t0)
                yield MutableAttr({'time': t, 'elevation': y + uniform(-noise, noise), 'distance': x})
                t += dt
            t0, x0, y0 = t1, x1, y1

    def test_build(self):
        waypoints = list(self.build_climb([(0, 0), (1000, 100), (2000, 0)]))
        # print(waypoints)
        self.assertEqual(len(waypoints), 2000 / 10 + 1)
        self.assertEqual(waypoints[0].distance, 0)
        self.assertEqual(waypoints[0].time, 0)
        self.assertEqual(waypoints[0].elevation, 0)
        self.assertEqual(waypoints[-1].distance, 2000)
        self.assertEqual(waypoints[-1].time, 200)
        self.assertEqual(waypoints[-1].elevation, 0)

    def test_single(self):
        waypoints = list(self.build_climb([(0, 0), (1000, 100), (2000, 0)]))
        # print(waypoints)
        c = list(find_climbs(waypoints))
        self.assertEqual(len(c), 1)
        self.assertEqual(c[0][0].time, 0)
        self.assertEqual(c[0][0].distance, 0)
        self.assertEqual(c[0][0].elevation, 0)
        self.assertEqual(c[0][1].time, 100)
        self.assertEqual(c[0][1].distance, 1000)
        self.assertEqual(c[0][1].elevation, 100)

    def test_noisy_single(self):
        for _ in range(10):
            # increase distance to peak to avoid cutoff at 1km
            waypoints = list(self.build_climb([(0, 0), (1100, 100), (2000, 0)], noise=1))
            # print(waypoints)
            c = list(find_climbs(waypoints))
            self.assertEqual(len(c), 1)
            self.assertAlmostEqual(c[0][0].time, 0, delta=5)
            self.assertAlmostEqual(c[0][0].distance, 0, delta=50)
            self.assertAlmostEqual(c[0][0].elevation, 0, delta=5)
            self.assertAlmostEqual(c[0][1].time, 110, delta=5)
            self.assertAlmostEqual(c[0][1].distance, 1100, delta=50)
            self.assertAlmostEqual(c[0][1].elevation, 100, delta=5)

    def test_multiple(self):
        waypoints = list(self.build_climb([(0, 0), (1100, 100), (1200, 90), (1500, 150)]))
        # print(waypoints)
        c = list(find_climbs(waypoints))
        # print(c)
        self.assertEqual(len(c), 1)
        self.assertEqual(c[0][1].elevation - c[0][0].elevation, 150)
        waypoints = list(self.build_climb([(0, 0), (1100, 100), (1200, 80), (1500, 150)]))
        # print(waypoints)
        c = list(find_climbs(waypoints))
        # print(c)
        self.assertEqual(len(c), 1)
        self.assertEqual(c[0][1].elevation - c[0][0].elevation, 100)
        waypoints = list(self.build_climb([(0, 0), (1100, 100), (1200, 80), (1500, 170)]))
        # print(waypoints)
        c = list(find_climbs(waypoints))
        # print(c)
        self.assertEqual(len(c), 2)
        self.assertEqual(c[0][1].elevation - c[0][0].elevation, 100)
        self.assertEqual(c[1][1].elevation - c[1][0].elevation, 90)


    # def test_noisy_single_to_failure(self):
    #     while True:
    #         waypoints = list(self.build_climb([(0, 0), (1100, 100), (2000, 0)], noise=1))
    #         c = list(climbs(waypoints))
    #         if len(c) != 1:
    #             climbs(waypoints)  # debug breakpoint here



def analyse_activity(log, s, id):
    aj = s.query(ActivityJournal).filter(ActivityJournal.id == id).one()
    waypoints = list(WaypointReader(log, with_timespan=False).read(s, aj, {DISTANCE: 'distance',
                                                                           ELEVATION: 'elevation'},
                                                                   SegmentImporter))
    plots = []
    for phi in (0.3, 0.4, 0.5, 0.6, 0.7, 1):
        # this is mainly noise getting things into appropriate forms for plotting...
        climbs = pd.DataFrame()
        for lo, hi in find_climbs(waypoints, params=Climb(phi=phi)):
            climbs = climbs.append(pd.Series({CLIMB_ELEVATION: hi.elevation - lo.elevation,
                                              CLIMB_DISTANCE: hi.distance - lo.distance}, name=hi.time))
        all = activity_statistics(s, DISTANCE, ELEVATION, activity_journal_id=id, with_timespan=True).dropna()
        all[DISTANCE_KM] = all[DISTANCE] / 1000
        all[ELEVATION_M] = all[ELEVATION]
        st = [df for id, df in all.groupby('timespan_id')]
        ys = [df[ELEVATION_M].copy() for df in st]
        for y, df in zip(ys, st):
            y.index = df[DISTANCE_KM].copy()
        f = line_diff_elevation_climbs(400, 200, ys, climbs=climbs, st=st)
        label = Label(x=220, y=100, x_units='screen', y_units='screen',
                      text='phi=%.1f (%d)' % (phi, len(climbs)), render_mode='css',
                      background_fill_color='white', background_fill_alpha=1.0)
        f.add_layout(label)
        plots.append(f)
    SingleShotServer(log, column(row(*plots[:2]), row(*plots[2:4]), row(*plots[4:])),
                     title='Climb Analysis', template=TEMPLATE, pause=5)


def analyse():
    log, db = config('-v 5')
    set_log(log)
    with db.session_context() as s:
        analyse_activity(log, s, _resolve_activity(s, '2017-09-29 16:30:00', None))
        analyse_activity(log, s, _resolve_activity(s, '2017-09-01 16:30:00', None))
        analyse_activity(log, s, _resolve_activity(s, '2017-03-28 16:30:00', None))


if __name__ == '__main__':
    analyse()




from json import dumps

from ch2.config import name_constant, add_enum_constant, set_constant
from ch2.stoats.calculate.climb import Climb

CLIMB_CNAME = 'Climb'


def add_climb(s, activity_group, phi=0.6):
    '''
    Add the constants necessary to auto-detect climbs.
    '''
    activity_group_constraint = str(activity_group)
    fatigue_name = name_constant(CLIMB_CNAME, activity_group)
    fatigue = add_enum_constant(s, fatigue_name, Climb, single=True, constraint=activity_group_constraint,
                                description='Data needed to cdetect climbs - see Climb enum')
    set_constant(s, fatigue, dumps({'phi': phi, 'min_gradient': 3, 'min_elevation': 80, 'max_reversal': 0.1}))




import datetime as dt

import numpy as np
import pandas as pd
from bokeh import palettes, tile_providers
from bokeh.models import NumeralTickFormatter, PrintfTickFormatter, Range1d, LinearAxis, PanTool, ZoomInTool, HoverTool, \
    ZoomOutTool, ResetTool
from bokeh.plotting import figure

from .data_frame import interpolate_to_index, delta_patches, closed_patch
from ..stoats.names import TIME, HR_ZONE, CLIMB_DISTANCE, CLIMB_ELEVATION


def clean(s):
    return s[~s.isin([np.nan, np.inf, -np.inf])]


def clean_all(ss):
    if ss is not None:
        return [clean(s) for s in ss]


def max_all(ss):
    return max(s.max() for s in ss if len(s))


def min_all(ss):
    return min(s.min() for s in ss if len(s))


def tools(x=None, y=None):
    tools = [PanTool(dimensions='width'),
             ZoomInTool(dimensions='width'), ZoomOutTool(dimensions='width'),
             ResetTool()]
    if x and y:
        tools.append(HoverTool(tooltips=[(x, '$x'), (y, '$y')]))
    return tools


def dot_map(n, x1, y1, size, x2=None, y2=None):

    f = figure(plot_width=n, plot_height=n, x_axis_type='mercator', y_axis_type='mercator',
               tools=[PanTool(), ZoomInTool(), ZoomOutTool(), ResetTool()])
    f.toolbar.logo = None

    for x, y, s in zip(x1, y1, size):
        f.circle(x=x, y=y, line_alpha=0, fill_color='red', size=s, fill_alpha=0.03)
        f.line(x=x, y=y, line_color='black')

    if x2 is not None:
        for x, y in zip(x2, y2):
            f.line(x=x, y=y, line_color='grey')

    f.add_tile(tile_providers.STAMEN_TERRAIN, alpha=0.1)
    f.axis.visible = False

    return f


def line_diff_elevation_climbs(nx, ny, y1, y2=None, climbs=None, st=None, y3=None, x_range=None):
    from .activity import DISTANCE_KM, ELEVATION_M
    f = line_diff(nx, ny, DISTANCE_KM, y1, y2=y2, x_range=x_range)
    if y3 is not None:
        for y in y3:
            f.line(x=y.index, y=y, color='black', alpha=0.1, line_width=2)
    if climbs is not None:
        all = pd.concat(st)
        for time, climb in climbs.iterrows():
            i = all.index.get_loc(time, method='nearest')
            x = all[DISTANCE_KM].iloc[i]
            x = (x - climb[CLIMB_DISTANCE] / 1000, x)
            y = all[ELEVATION_M].iloc[i]
            y = (y - climb[CLIMB_ELEVATION], y)
            f.line(x=x, y=y, color='red', line_width=5, alpha=0.2)
            for xx, yy in zip(x, y):
                f.circle(x=xx, y=yy, color='red', size=8, alpha=0.2)
    return f


def line_diff(nx, ny, xlabel, y1, y2=None, x_range=None):

    y1, y2 = clean_all(y1), clean_all(y2)
    is_x_time = any(isinstance(s.index[0], dt.datetime) for s in y1 if len(s))

    f = figure(plot_width=nx, plot_height=ny, x_axis_type='datetime' if is_x_time else 'linear',
               tools=tools(xlabel, y1[0].name))
    f.toolbar.logo = None

    y_min, y_max = min_all(y1), max_all(y1)
    if y2:
        y_min, y_max = min(y_min, min_all(y2)), max(y_max, max_all(y2))
    dy = y_max - y_min

    if is_x_time:
        f.xaxis.axis_label = 'Time'
        f.xaxis[0].formatter = NumeralTickFormatter(format='00:00:00')
        zero = min(y.index.min() for y in y1)
        for y in y1:
            y.index = (y.index - zero).total_seconds()
    else:
        f.xaxis.axis_label = xlabel
        f.xaxis[0].formatter = PrintfTickFormatter(format='%.2f')
    f.yaxis.axis_label = y1[0].name

    f.y_range = Range1d(start=0 if y_min == 0 else y_min - 0.1 * dy, end=y_max + 0.1 * dy)
    for y in y1:
        f.line(x=y.index, y=y, color='black')

    if y2:
        if is_x_time:
            zero = min(y.index.min() for y in y2)
            for y in y2:
                y.index = (y.index - zero).total_seconds()
        for y in y2:
            f.line(x=y.index, y=y, color='grey')

        y1, y2 = pd.concat(y1), pd.concat(y2)
        y2 = interpolate_to_index(y1, y2)
        y1, y2, range = delta_patches(y1, y2)
        f.extra_y_ranges = {'delta': range}
        if y1 is not None:
            f.patch(x=y1.index, y=y1, color='green', alpha=0.1, y_range_name='delta')
        if y2 is not None:
            f.patch(x=y2.index, y=y2, color='red', alpha=0.1, y_range_name='delta')

    # f.toolbar_location = None
    if x_range is not None:
        f.x_range = x_range

    return f


def cumulative(nx, ny, y1, y2=None, sample=10):

    y1 = pd.concat(y1)
    y1 = y1.sort_values(ascending=False).reset_index(drop=True)
    y_max = y1.max()
    y_min = y1.min()
    if y2 is not None:
        y2 = pd.concat(y2)
        y2 = y2.sort_values(ascending=False).reset_index(drop=True)
        y_max = max(y_max, y2.max())
        y_min = min(y_min, y2.min())
    dy = y_max - y_min

    f = figure(plot_width=nx, plot_height=ny,
               x_range=Range1d(start=y1.index.max() * sample, end=0),
               x_axis_type='datetime',
               x_axis_label=TIME,
               y_range=Range1d(start=0 if y_min == 0 else y_min - 0.1 * dy, end=y_max + 0.1 * dy),
               y_axis_location='right',
               y_axis_label=y1.name)
    f.xaxis[0].formatter = NumeralTickFormatter(format='00:00:00')

    f.line(x=y1.index * sample, y=y1, color='black')
    if y2 is not None:
        f.line(x=y2.index * sample, y=y2, color='grey')
        y1, y2, range = delta_patches(y1, y2)
        f.extra_y_ranges = {'delta': range}
        f.patch(x=y1.index * sample, y=y1, color='green', alpha=0.1, y_range_name='delta')
        f.patch(x=y2.index * sample, y=y2, color='red', alpha=0.1, y_range_name='delta')

    f.toolbar_location = None
    return f


def health(nx, ny, ftn, ftg, hr):

    f = figure(plot_width=nx, plot_height=ny, x_axis_type='datetime', tools=tools())
    f.toolbar.logo = None
    f.xaxis.axis_label = 'Date'

    max_f = ftn.max() * 1.1
    min_f = ftn.min() * 0.9

    f.y_range = Range1d(start=min_f, end=max_f)
    f.yaxis.axis_label = '%s, %s' % (ftn.name, ftg.name)
    f.yaxis[0].formatter = PrintfTickFormatter(format='')

    patch = closed_patch(ftg, zero=min_f)
    f.patch(x=patch.index, y=patch, color='grey', alpha=0.2)
    f.line(x=ftn.index, y=ftn, color='black')

    hr = hr.dropna()
    if len(hr):
        max_hr = hr.max() * 1.1
        min_hr = hr.min() * 0.9
        f.extra_y_ranges = {hr.name: Range1d(start=min_hr, end=max_hr)}
        f.add_layout(LinearAxis(y_range_name=hr.name, axis_label=hr.name), 'right')
        f.circle(x=hr.index, y=hr, color='red', alpha=0.2, y_range_name=hr.name)

    # f.toolbar_location = None
    return f


def activities(nx, ny, steps, active_time):

    f = figure(plot_width=nx, plot_height=ny, x_axis_type='datetime', tools=tools())
    f.toolbar.logo = None
    f.xaxis.axis_label = 'Date'

    steps = None if steps is None else steps.dropna()
    active_time = None if active_time is None else active_time.dropna()

    if steps is not None and len(steps):
        f.y_range = Range1d(start=0, end=1.1 * steps.max())
        f.yaxis.axis_label = steps.name
        f.vbar(x=steps.index, width=dt.timedelta(hours=20), top=steps, fill_color='grey', fill_alpha=0.4, line_alpha=0)
    else:
        f.yaxis[0].visible = False

    if active_time is not None and len(active_time):
        f.extra_y_ranges = {active_time.name: Range1d(start=0, end=active_time.max() * 1.1)}
        f.add_layout(LinearAxis(y_range_name=active_time.name, axis_label=active_time.name), 'right')
        f.circle(x=active_time.index, y=active_time, color='black', fill_alpha=0, y_range_name=active_time.name)
        f.yaxis[1].formatter = PrintfTickFormatter(format='')

    # f.toolbar_location = None
    return f


def heart_rate_zones(nx, ny, hrz):

    max_z, n_sub = 5, 5

    bins = pd.interval_range(start=1, end=max_z+1, periods=n_sub * max_z, closed='left')
    c = [palettes.Inferno[6][int(b.left)-1] for b in bins]
    hrz_categorized = pd.cut(hrz, bins)
    counts = hrz_categorized.groupby(hrz_categorized).count()

    f = figure(plot_width=nx, plot_height=ny, x_range=Range1d(start=1, end=max_z+1), x_axis_label=HR_ZONE)
    f.quad(left=counts.index.categories.left, right=counts.index.categories.right, top=counts, bottom=0,
           color=c, fill_alpha=0.2)
    f.yaxis.visible = False

    f.toolbar_location = None
    return f

